import time
import datetime
import re
from enum import IntEnum


import pymfx4.mfdatastorage as ds
import pymfx4.mfparameters as mfp
import pymfx4.mffullindex as mff
from pymfx4.mfkeyboard import keyboard as kb
from pymfx4 import mfl


class mfx4device:
    """Base class for the MFX_4 devices family.

    Contains wrappers for base function
    - Determining the device class (edi / terminal / controller) and version.
    - reboot device, save parameters, write/read parameters.
    """

    IxCutSaveAllParameter = mff.FullIndex(0, 0x1010, 1)
    IxCutCPUClock = mff.FullIndex(0, 0x200B, 0x11)
    IxUploadBin = mff.FullIndex(0, 0x200B, 16)
    IxUpload = mff.FullIndex(0, 0x200B, 8)
    IxUploadAck = mff.FullIndex(0, 0x200B, 0x07)
    IxUploadState = mff.FullIndex(0, 0x200B, 0x09)
    IxUploadCRC = mff.FullIndex(0, 0x200B, 13)
    IxSetMode = mff.FullIndex(0, 0x200B, 5)

    IxManufacturerDeviceName = mff.FullIndex(0, 0x1008, 0x00)
    IxManufacturerSoftwareVersion = mff.FullIndex(0, 0x100A, 0x00)
    IxManufacturerSoftwareDate = mff.FullIndex(0, 0x100A, 0x01)
    IxManufacturerSoftwareTime = mff.FullIndex(0, 0x100A, 0x02)

    IxStoreParameterSaveAll = mff.FullIndex(0, 0x1010, 0x01)
    IxRestoreDefaultParametersIx = mff.FullIndex(0, 0x1011, 3)

    IxAliveTime = mff.FullIndex(0, 0x200B, 14)

    # device soft reset
    DEVICE_REBOOT = 0x64616F6F
    DEVICE_SAVE_ALL_PARAMETERS = 0x65766173

    def __init__(self, NodeId):
        self._Node = NodeId
        ds.storage.AddGenericDevice(NodeId)
        self.GetInfo()

    def TheNode(self):
        return self._Node

    DeviceName = ds.dsproperty(Address=IxManufacturerDeviceName, NodeId=TheNode)
    SoftwareVersion = ds.dsproperty(Address=IxManufacturerSoftwareVersion, NodeId=TheNode)
    SoftwareDate = ds.dsproperty(Address=IxManufacturerSoftwareDate, NodeId=TheNode)
    SoftwareTime = ds.dsproperty(Address=IxManufacturerSoftwareTime, NodeId=TheNode)

    IxStoreParameterSaveAll = ds.dsproperty(Address=IxStoreParameterSaveAll, NodeId=TheNode)

    def GetInfo(self):
        ds.storage.Get(self.DeviceName)
        time.sleep(0.1)
        ds.storage.Get(self.DeviceName)
        time.sleep(0.1)
        ds.storage.Get(self.SoftwareVersion)
        time.sleep(0.1)
        ds.storage.Get(self.SoftwareDate)
        time.sleep(0.1)
        ds.storage.Get(self.SoftwareTime)
        time.sleep(0.1)

    @property
    def Version(self):
        such = self.SoftwareVersion.Value
        if such is None:
            such = ""
        return list(map(int, re.findall(r'([0-9]+)[.]{0,1}', such, re.M | re.I)))

    @property
    def BuildDate(self):
        software_date = self.SoftwareDate.Value
        software_time = self.SoftwareTime.Value
        if software_time is None:
            software_time = ""
        if software_date is None:
            software_date = ""
        try:
            combine_f1 = datetime.datetime.strptime(software_date.strip(), '%b %d %Y')
            combine_f2 = datetime.datetime.strptime(software_time.strip(), '%H:%M:%S').time()
            return datetime.datetime.combine(combine_f1, combine_f2)
        except Exception as ex:
            pass
        return datetime.datetime(1900, 1, 1, 0, 0, 0, 0)

    @property
    def FullVersion(self):
        return (self.Version, self.BuildDate)

    def SaveParameters(self):
        ds.storage.Get(self.SoftwareTime)
        ds.storage.Set(self.IxStoreParameterSaveAll, mfx4device.DEVICE_SAVE_ALL_PARAMETERS)

    def Reboot(self):
        ds.storage.Set(self.IxStoreParameterSaveAll, mfx4device.DEVICE_REBOOT)

    def PC1Reset(self):
        ds.storage.Set(mfx4Controller.IxProcessControlCommand.Node(self.TheNode()), 2)
        time.sleep(8)

    def Set(self, FI, Value):
        ds.storage.Set(FI.Node(self.TheNode()), Value)

    def Get(self, FI):
        FI = FI.Node(self.TheNode())
        ds.storage.Get(FI)
        time.sleep(0.8)
        return ds.storage.GetNative(FI)
    
    def SetNotification(self, FI, Value, Func):
        ds.storage.SetNotification(FI.Node(self.TheNode()), Value, Func)

    def RemoveNotification(self, FI):
        ds.storage.RemoveNotification(FI.Node(self.TheNode()))

    def DeployAndCheckEntrywise(self, package, iter=4):
        if isinstance(package, mfp.mfparameters):
            for i in range(iter):
                if mfp.mfParameterManager.DeployAndCheckEntrywise(package, self):
                    return True
                mfl.info("Check failed! Trying again.")
                time.sleep(1)
        return False

    def DeployAndCheck(self, package, iter=4):
        if isinstance(package, mfp.mfparameters):
            for i in range(iter):
                if mfp.mfParameterManager.DeployAndCheck(package, self):
                    return True
                mfl.info("Check failed! Trying again.")
                time.sleep(1)
        return False

    def __eq__(self, other):
        if not isinstance(other, mfx4device):
            return NotImplemented
        return self == other

    def __repr__(self):
        if self.BuildDate != datetime.datetime(1900, 1, 1, 0, 0, 0, 0):
            return f'{self.SoftwareVersion.Value} - N{self._Node:00} - {self.BuildDate:yyyy-mm-dd HH:MM:SS}'
        return f'{self.SoftwareVersion.Value} - N{self._Node:00}'


class AccessControlLevel_t(IntEnum):
    AccessControl_FactorySettings = 0
    AccessControl_Calibration = 1
    AccessControl_Service = 2
    AccessControl_Manager = 3
    AccessControl_Operator = 4
    AccessControl_None = 7
    AccessControl_Default = AccessControl_None


class mfx4Controller(mfx4device):
    """Represents an MFX_4 Controller V2.

    Used to distinfffguish devices.
    Contains relevant controller CANOpen addresses.
    """

    IxAccessRightsActGroup = mff.FullIndex(0, 0x2820, 0x01)
    IxAccessRightsKey = mff.FullIndex(0, 0x2820, 0x02)
    IxAccessRightsLogin = mff.FullIndex(0, 0x2820, 0x03)
    IxAccessRightsPWService = mff.FullIndex(0, 0x2820, 0x04)
    IxAccessRightsPWManager = mff.FullIndex(0, 0x2820, 0x05)
    IxAccessRightsPWOperator = mff.FullIndex(0, 0x2820, 0x06)
    # IxAccessRightsCalibrationMode = mff.FullIndex(0,0x2820,0x06)

    # Controller specific
    IxProcessControlCommand = mff.FullIndex(0, 0x2425, 11)
    IxProcessControlTransaction = mff.FullIndex(0, 0x2425, 0x0D)
    IxProcessControlBatch = mff.FullIndex(0, 0x2425, 0x0C)
    IxDigitalInputBits = mff.FullIndex(0, 0x2705, 0x0)
    IxDigitalOutputBits = mff.FullIndex(0, 0x2707, 0x0)

    IxPulseAB_Mode = mff.FullIndex(0, 0x21F0, 0x01)
    IxPulseAB_MaxFrequency = mff.FullIndex(0, 0x21F0, 0x04)
    IxPulseAB_Frequency = mff.FullIndex(0, 0x21F0, 0x07)
    IxPulseAB_PulseLost = mff.FullIndex(0, 0x21F0, 0x0A)
    IxPulseAB_PulseRevers = mff.FullIndex(0, 0x21F0, 0x09)
    IxPulseAB_PulseAll = mff.FullIndex(0, 0x21F0, 0x0B)
    IxPulseAB_PulseA = mff.FullIndex(0, 0x21F0, 0x0C)

    IxPulseC_Mode = mff.FullIndex(0, 0x21F1, 0x01)
    IxPulseC_MaxFrequency = mff.FullIndex(0, 0x21F1, 0x04)
    IxPulseC_Frequency = mff.FullIndex(0, 0x21F1, 0x07)
    IxPulseC_Pulse = mff.FullIndex(0, 0x21F1, 0x0C)

    IxPulseD_Mode = mff.FullIndex(0, 0x21F2, 0x01)
    IxPulseD_MaxFrequency = mff.FullIndex(0, 0x21F2, 0x04)
    IxPulseD_Frequency = mff.FullIndex(0, 0x21F2, 0x07)
    IxPulseD_PulseAll = mff.FullIndex(0, 0x21F2, 0x0B)

    IxPressure1_Mode = mff.FullIndex(0, 0x2620, 0x01)
    IxPressure1_PressureLow = mff.FullIndex(0, 0x2620, 0x02)
    IxPressure1_PressureHigh = mff.FullIndex(0, 0x2620, 0x03)
    IxPressure1_AnalogLow = mff.FullIndex(0, 0x2620, 0x04)
    IxPressure1_AnalogHigh = mff.FullIndex(0, 0x2620, 0x05)
    IxPressure1_PressureMinError = mff.FullIndex(0, 0x2620, 0x06)
    IxPressure1_PressureMaxError = mff.FullIndex(0, 0x2620, 0x07)
    IxPressure1_PressureShiftError = mff.FullIndex(0, 0x2620, 0x08)
    IxPressure1_PressureValue = mff.FullIndex(0, 0x2620, 0x09)
    IxPressure2_Mode = mff.FullIndex(0, 0x2621, 0x01)
    IxPressure2_PressureLow = mff.FullIndex(0, 0x2621, 0x02)
    IxPressure2_PressureHigh = mff.FullIndex(0, 0x2621, 0x03)
    IxPressure2_AnalogLow = mff.FullIndex(0, 0x2621, 0x04)
    IxPressure2_AnalogHigh = mff.FullIndex(0, 0x2621, 0x05)
    IxPressure2_PressureMinError = mff.FullIndex(0, 0x2621, 0x06)
    IxPressure2_PressureMaxError = mff.FullIndex(0, 0x2621, 0x07)
    IxPressure2_PressureShiftError = mff.FullIndex(0, 0x2621, 0x08)
    IxPressure2_PressureValue = mff.FullIndex(0, 0x2621, 0x09)

    IxError1_LastError = mff.FullIndex(0, 0x2805, 0x07)
    IxError1_ErrorNo0 = mff.FullIndex(0, 0x2805, 0x09)
    IxError1_ErrorNo1 = mff.FullIndex(0, 0x2805, 0x0A)

    IxExtLog_Blk1 = mff.FullIndex(0, 0x2A40, 0x03)
    IxExtLog_Blk2 = mff.FullIndex(0, 0x2A40, 0x04)
    IxExtLog_Blk3 = mff.FullIndex(0, 0x2A40, 0x05)
    IxExtLog_Blk4 = mff.FullIndex(0, 0x2A40, 0x06)
    IxExtLog_Blk5 = mff.FullIndex(0, 0x2A40, 0x07)
    IxExtLog_Blk6 = mff.FullIndex(0, 0x2A40, 0x08)
    IxExtLog_Blk7 = mff.FullIndex(0, 0x2A40, 0x09)
    IxExtLog_Blk8 = mff.FullIndex(0, 0x2A40, 0x0A)
    IxExtLog_Blk9 = mff.FullIndex(0, 0x2A40, 0x0B)
    IxExtLog_Blk10 = mff.FullIndex(0, 0x2A40, 0x0C)
    IxExtLog_Blk11 = mff.FullIndex(0, 0x2A40, 0x0D)

    ConstAccessControlMultiplier = 0x1000000
    ConstAccessControlFactory = 4712

    def __init__(self, NodeId):
        self._Node = NodeId
        ds.storage.AddGenericDevice(NodeId)
        self.GetInfo()

    def TheNode(self):
        return self._Node


class mfx4Cut(mfx4device):
    """Base class representing representing MFX_4 CUT devices EDI / TERMINAL.

    Used to distinfffguish devices.
    Contains relevant CUT CANOpen addresses.
    """

    IxMeter1State = mff.FullIndex(0, 0x2200, 0x09)

    def __init__(self, NodeId):
        self._Node = NodeId
        ds.storage.AddGenericDevice(NodeId)
        self.GetInfo()

    def TheNode(self):
        return self._Node


class mfx4Terminal(mfx4Cut):

    IxLastKeySimulator = mff.FullIndex(0, 0x2A60, 2)
    IxLastKey = mff.FullIndex(0, 0x200B, 6)

    IxTerminalTerminalInputState = mff.FullIndex(0, 0x2300, 8)
    IxTerminalState = mff.FullIndex(0, 0x2300, 7)
    IxTerminalMeterIndex = mff.FullIndex(0, 0x2300, 9)

    IxBOLConsistency = mff.FullIndex(0, 0x2450, 0x16)
    IxBolPrintingState = mff.FullIndex(0, 0x2450, 0x0A)
    IxBOLPrintingMeter = mff.FullIndex(0, 0x2450, 0x0B)

    IxBOLConsistency = mff.FullIndex(0, 0x2450, 0x16)
    IxBolPrintingState = mff.FullIndex(0, 0x2450, 0x0A)
    IxBOLPrintingMeter = mff.FullIndex(0, 0x2450, 0x0B)

    IxCardreaderMulti1State = mff.FullIndex(0, 0x2090, 0x04)

    IxDisplayObjectCharset = mff.FullIndex(0, 0x2005, 0x0C)

    IxHintTextMode = mff.FullIndex(0, 0x2006, 0x03)
    IxHintTextNumber = mff.FullIndex(0, 0x2006, 0x04)
    IxHintText1 = mff.FullIndex(0, 0x2006, 0x06)
    IxHintText2 = mff.FullIndex(0, 0x2006, 0x07)
    IxHintText3 = mff.FullIndex(0, 0x2006, 0x08)
    IxHintText4 = mff.FullIndex(0, 0x2006, 0x09)
    IxHintText5 = mff.FullIndex(0, 0x2006, 0x0A)
    IxHintText6 = mff.FullIndex(0, 0x2006, 0x0B)
    IxHintText7 = mff.FullIndex(0, 0x2006, 0x0C)
    IxHintText8 = mff.FullIndex(0, 0x2006, 0x0D)
    IxHintText9 = mff.FullIndex(0, 0x2006, 0x0E)
    IxHintText10 = mff.FullIndex(0, 0x2006, 0x0F)

    IxArmDefinition = mff.FullIndex(0, 0x2911, 0x03)

    IxMeterSpecific1DisplayLines = mff.FullIndex(0, 0x2901, 0x0D)

    IxRecipe4 = mff.FullIndex(0, 0x2911, 0x07)
    IxRecipe5 = mff.FullIndex(0, 0x2911, 0x08)

    def __init__(self, NodeId):
        self._Node = NodeId
        ds.storage.AddGenericDevice(NodeId)
        self.GetInfo()

    def TheNode(self):
        return self._Node

    def pressKey(self, key, sleep=1):
        ds.storage.Set(mfx4Terminal.IxLastKeySimulator.Node(self._Node), key)
        time.sleep(sleep)

    def pressKeys(self, keylist):
        for key in keylist:
            self.pressKey(key)

    def pressKeysFast(self, keylist):
        for key in keylist:
            self.pressKey(key, 0.2)

    def DelInput(self):
        for i in range(20):
            self.pressKey(kb.DEL, 0.1)

    def Home(self):
        print("Zurueck zur frmArmliste")
        self.pressKeysFast([kb.F1, kb.F1])

    def PresetArm(self, ArmNr):
        print("Preset Arm " + str(ArmNr))
        self.pressKeysFast([kb.Numbers[ArmNr], kb.F3])
        self.pressKeysFast([kb.Zwei, kb.Null, kb.Null, kb.Null, kb.F4])
        self.pressKeysFast([kb.Eins, kb.F4])
        time.sleep(5)
        self.pressKeysFast([kb.F3])

    def StopArm(self, ArmNr):
        print("Stop Arm " + str(ArmNr))
        self.pressKeysFast([kb.Numbers[ArmNr], kb.F3])

    def PrintArm(self, ArmNr):
        print("Print Arm " + str(ArmNr))
        self.pressKeysFast([kb.Numbers[ArmNr], kb.F2])
        self.pressKey(kb.Numbers[7])


def GetDeviceWithType(deviceType, deviceArray):
    for dev in deviceArray:
        if isinstance(dev, deviceType):
            return dev
    return None



"""Provides convenient functions to connect to MFX_4 devices.

This file contains the following functions:
    * mfx_endpoints - creates the connection, returns a tuple of MFX_4 device Endpoints.
    * mfx_close_endpoints - closes the connection
"""

import pymfx4.mfdatastorage as ds
from pymfx4.mffullindex import FullIndex
from pymfx4 import mfl

import socket
import serial
import time
import threading
import re
import sys
import traceback



class mfx4ExceptionEndpointDirect(Exception):
    pass


class mfx4ExceptionEndpointRemote(Exception):
    pass


class mfconnection:
    """Creates a connection to an MFX_4 device.

    Creates a connection (primarily socket TCP-IP) with an MFX_4 device and
    starts a read-thread to interpret the incoming MFX_4 protocol messages.
    Appropriately wraps (MFX_4 protocol) and writes bytes/strings into the socket.
    This class is not used directly, see mfx_endpoints and mfx_close_endpoints.
    """

    TCPIP_STR = "TCPIP"
    SERIAL_STR = "SERIAL"
    DUMMY_STR = "DUMMY"

    def __init__(self, config, tim=5):
        self.config = config
        self.TheEncoding = "ascii"
        self.sock = None
        self.connectionType = None
        self.CloseThreads = False
        self.timeouts = 0
        self.timeout = tim
        self.dataBuf = bytearray()
        self.SocketException = False

    @property
    def IsDummy(self):
        return self.connectionType == mfconnection.DUMMY_STR

    def _read_from_port(self):
        start = time.time()
        while True:
            if self.CloseThreads:
                return
            try:
                data = self.Read(512)
                self.dataBuf.extend(data)
                # print("Data rcvd: ", len(data))
                # print(f"HOST <-{data}")
                if len(self.dataBuf) > 0 and b'\x60' in self.dataBuf:

                    mfl.debug(f"HOST <-Log_Encoded: {data.decode(self.TheEncoding, 'ignore')}")
                    mfl.debug(f"HOST <-Log_Raw: {data}")

                    messageStream = self.dataBuf.decode(self.TheEncoding, "ignore")
                    messages = messageStream.split("`")

                    self.dataBuf = bytearray()
                    if not messageStream.endswith("`"):
                        self.dataBuf.extend(messages[-1])
                        messages.remove(messages[-1])

                    for line in messages:
                        # print(line)
                        matchObj = re.match(r'(?P<cmd>.*?)_(?P<nodeid>.*?)_(?P<index>.*?)_(?P<subindex>.*?)_(?P<content>.*)', line, re.I | re.DOTALL)
                        if matchObj:
                            try:
                                fi = FullIndex(int(matchObj.group("nodeid")), int(matchObj.group("index")), int(matchObj.group("subindex")))
                                # print(fi)
                                # print(matchObj.group("content"))
                                content = matchObj.group("content")
                                ds.storage.Notify(fi, content)
                            except Exception as e:
                                try:
                                    mfl.critical(f"Exception: {e}")
                                    mfl.critical(f"Exception Occured; {traceback.format_exception(*sys.exc_info())}")
                                except Exception as ex2:
                                    raise Exception('Thread unknown exception occured.')
                theTime = time.time() - start
                if theTime > self.timeout:
                    if self.connectionType == mfconnection.TCPIP_STR:
                        self.SendBytes(bytes("`", self.TheEncoding, errors='strict'))
                    start = time.time()
            except Exception as ex:
                if self.connectionType == mfconnection.TCPIP_STR:
                    self.SendBytes(bytes("`", self.TheEncoding, errors='strict'))
                self.SocketException = True

    def Send(self, befehl):
        bf = bytes(befehl, self.TheEncoding, errors='strict')
        mfl.debug(f"HOST-> {befehl}")

        if self.connectionType == mfconnection.SERIAL_STR:
            self.sock.write(bf)
        elif self.connectionType == mfconnection.TCPIP_STR:
            try:
                self.sock.send(bf)
                self.timeouts = 0
            except socket.timeout:
                mfl.info("Send timed out.")
                self.timeouts = self.timeouts + 1
            except Exception as E:
                self.SocketException = True

    def SendBytes(self, theBytes):
        if self.connectionType == mfconnection.SERIAL_STR:
            self.sock.write(theBytes)
        elif self.connectionType == mfconnection.TCPIP_STR:
            try:
                self.sock.send(theBytes)
                self.timeouts = 0
            except socket.timeout:
                mfl.info("Send timed out.")
                self.timeouts += 1
            except Exception as E:
                self.SocketException = True

    def Read(self, bytes):
        rec = ""
        if self.connectionType == mfconnection.SERIAL_STR:
            rec = self.sock.read(bytes)
        elif self.connectionType == mfconnection.TCPIP_STR:
            rec = self.sock.recv(bytes,)
        return rec

    def close(self):
        try:
            if self.IsDummy:
                pass
            else:
                self.sock.close()
        except Exception as ex:
            pass

    def start_thread(self):
        if self.IsDummy:
            pass
        self.CloseThreads = False
        threading.Thread(target=self._read_from_port, name="ReadThread").start()
        # time.sleep(1)

    def close_thread(self):
        if self.IsDummy:
            pass
        self.CloseThreads = True
        time.sleep(5)

    def open(self):
        if "dummy" in self.config.lower():
            self.connectionType = mfconnection.DUMMY_STR
        elif "com" in self.config.lower():
            self.connectionType = mfconnection.SERIAL_STR
            address, baudrate, DATASIZE, PAR2, STOP = self.config.split(":")

            PAR = serial.PARITY_NONE
            if PAR2 == "E":
                PAR = serial.PARITY_EVEN
            elif PAR2 == "O":
                PAR = serial.PARITY_ODD

            self.sock = serial.Serial(address.lower(), int(baudrate), timeout=self.timeout, parity=PAR, bytesize=int(DATASIZE), stopbits=int(STOP))
        else:
            address, port = self.config.split(":")
            if not port:
                port = "4003"
            socket.setdefaulttimeout(None)
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((address, int(port)))
            self.sock.send(bytes("`", self.TheEncoding, errors='strict'))
            self.connectionType = mfconnection.TCPIP_STR

    def open_and_start_thread(self):
        try:
            self.open()
            # time.sleep(0.1)
            self.start_thread()
            self.SocketException = False
            return True
        except Exception as Ex:
            mfl.critical(traceback.format_exc())
            pass
        return False

    def close_and_end_thread(self):
        try:
            self.close_thread()
            self.close()
            return True
        except Exception as ex:
            pass
        return False


def mfx_endpoints(mfx_connDevice, mfx_EndpointNodes):
    """Creates a connection with MFX_4 devices.

    Creates either a socket tcp/ip or serial connection with <mfx_connDevice>.
    Creates a read-thread to interpret the incoming MFX_4 protocol messages.
    Determines whether the listed <mfx_EndpointNodes> are online and adds them
    to the datastorage.

    TODO make more than one connection possible with this scheme.

    Args:
        mfx_connDevice (str): Create a connection to this device. E.g. 192.168.3.11:4004 or 9600:8:N:1
        mfx_EndpointNodes (list): List of node-ids that are to be expected.

    Returns:
        tuple: a tuple of objects in their relevant class. E.g. a mfx4Controller-instance
            is returned for each controller listed in <mfx_EndpointNodes>.

    Raises:
        mfx4ExceptionEndpointDirect: If no MFX_4 device can be determined through <mfx_connDevice>.
        mfx4ExceptionEndpointRemote: At least one listed endpoint in <mfx_EndpointNodes> can not be found.
    """

    from pymfx4.mfconnection import mfconnection
    import pymfx4.mfdatastorage as ds
    from pymfx4.mfx4ARMadillo import mfx4ARMadillo
    from pymfx4.mfx4device import mfx4device
    from pymfx4.mfx4device import mfx4Controller
    from pymfx4.mfx4device import mfx4Terminal
    from pymfx4.mfx4device import mfx4Cut
    from pymfx4 import mfl
    import time

    def DetermineDeviceTypeBy(Node):
        device = mfx4device(Node)
        time.sleep(0.5)
        for i in range(4):
            if device.SoftwareVersion.Value is not None:
                break
            device.GetInfo()
            time.sleep(0.5)

        firmware = ""
        if device.SoftwareVersion.Value is not None:
            firmware = device.SoftwareVersion.Value.lower()
        if firmware.find('ctl') >= 0:
            version = device.Version
            if version is not None:
                if version[0] > 2:
                    mfl.info("Node '{Node}' is mfx4ARMadillo.")
                    device = mfx4ARMadillo(Node)
                else:
                    mfl.info(f"Node '{Node}' is mfx4Controller.")
                    device = mfx4Controller(Node)
                time.sleep(0.5)
                for i in range(4):
                    if device.SoftwareVersion.Value is not None:
                        break
                    device.GetInfo()
                    time.sleep(0.5)

        elif firmware.find('term') >= 0 or firmware.find('cut') >= 0 or firmware.find('edi') >= 0:
            if firmware.find('term') >= 0:
                mfl.info(f"Node '{Node}' is mfx4Terminal.")
                device = mfx4Terminal(Node)
            else:
                mfl.info(f"Node '{Node}' is mfx4CUT.")
                device = mfx4Cut(Node)
            time.sleep(0.5)
            for i in range(4):
                if device.SoftwareVersion.Value is not None:
                    break
                device.GetInfo()
                time.sleep(0.5)

        return device

    # CONNECTION
    mfc = mfconnection(mfx_connDevice)
    ds.storage.mfc = mfc
    mfc.open_and_start_thread()

    # Fill Datastorage with correct eds/xdd
    devices = []

    ConnDevice = DetermineDeviceTypeBy(0)
    # time.sleep(1)
    if ds.storage.AddDevice(ConnDevice) is False:
        mfc.close_and_end_thread()
        mfl.error("Could not determine device-type (direct)")
        raise mfx4ExceptionEndpointDirect("Could not determine device-type (direct)")

    devices.append(ConnDevice)

    # Endpoint = ConnDevice

    N = mfx_EndpointNodes
    if not isinstance(N, list):
        N = [N]

    for node in N:
        if isinstance(node, str):
            node = int(node)
        if node > 0:
            Endpoint = DetermineDeviceTypeBy(node)
            if ds.storage.AddDevice(Endpoint) is False:
                mfc.close_and_end_thread()
                mfl.error("Could not determine device-type (remote)")
                raise mfx4ExceptionEndpointRemote("Could not determine device-type (remote)")

            devices.append(Endpoint)

    return tuple(devices)


def mfx_close_endpoints():
    """Closes the connection opened via mfx_endpoints.

    TODO make more than one connection possible with this scheme.
    """
    import pymfx4.mfdatastorage as ds
    if ds.storage.mfc:
        ds.storage.mfc.close_and_end_thread()



class FullIndex(object):
    """Represents a CANOpen address with node-id, index and subindex."""
    
    def __init__(self, node, index, subindex):
        self.NodeId = node
        self.Index = index
        self.SubIndex = subindex

    def __hash__(self):
        i = self.NodeId << 24 | self.Index << 8 | self.SubIndex
        return hash(i)

    def __eq__(self, other):
        # return (self.NodeId, self.Index, self.SubIndex) == (other.Node, other.Index, other.SubIndex)
        return self.__hash__() == other.__hash__()

    def __ne__(self, other):
        # Not strictly necessary, but to avoid having both x==y and x!=y
        # True at the same time
        return not (self == other)

    def __getitem__(self, k):
        return k

    def __repr__(self):
        return 'N{0:00}-0x{1:4X}-0x{2:02X}'.format(self.NodeId, self.Index, self.SubIndex)

    def Node(self, newnode):
        # if isinstance(newnode, mfx4device):
        #    return FullIndex(newnode.TheNode(),self.Index,self.SubIndex)
        return FullIndex(newnode, self.Index, self.SubIndex)



# -*- coding: utf-8 -*-

from .context import pymfx4

import unittest


class AdvancedTestSuite(unittest.TestCase):
    """Advanced test cases."""

    def test_error_binary_unpack(self):
        pass

if __name__ == '__main__':
    unittest.main()



"""Defines and instanciates the mfDatastorage.

Attributes:
    storage (mfDatastorage): Central piece of CANOpen datamanagement within the pymfx4 package.

Todo:
    * Allow for more than one connection at the same time.
"""

from pymfx4.mffullindex import FullIndex
import pymfx4.mfx4driver as mfx
from pymfx4.mfxdd import genericXDD, cutXDD, controllerXDD
from pymfx4 import mfl
import xml.etree.ElementTree as ET
import base64

# class CanOpenObject(object):
#     """Represents a CANOpen Sub/Object."""

#     def __init__(self, name, objectType, dataType, lowLimit, highLimit, defaultValue, accessType, PDOmapping):
#         self.name = name
#         self.objectType = objectType
#         self.dataType = dataType
#         self.lowLimit = lowLimit
#         self.highLimit = highLimit
#         self.defaultValue = defaultValue
#         self.accessType = accessType
#         self.PDOmapping = PDOmapping


class DatastorageEntry(object):
    """Represents an CANOpen entry."""

    datatype_convert = {
        -1: "c",
        0x01: "c",
        0x02: "c",
        0x03: "i",
        0x04: "I",
        0x05: "C",
        0x06: "I",
        0x07: "L",
        0x08: "F",
        0x09: "S",
        0x0E: "d",
        0x0F: "S",
    }
    # BOOLEAN         = 0x0001,
    # INTEGER8        = 0x0002,
    # INTEGER16       = 0x0003,
    # INTEGER32       = 0x0004,
    # UNSIGNED8       = 0x0005,
    # UNSIGNED16      = 0x0006,
    # UNSIGNED32      = 0x0007,
    # REAL32          = 0x0008,
    # VISIBLE_STRING  = 0x0009,
    # REAL64          = 0x000E, // BIT_STRING, used for double values
    # //reserved        = 0x000E, // official CANopen spec.
    # DOMAIN          = 0x000F,

    NumToTypeFunc = {

        0x01: bool,

        # signed
        0x02: int,
        0x03: int,
        0x04: int,

        # unsigned
        0x05: int,
        0x06: int,
        0x07: int,

        # floating point
        0x08: float,  # float
        0x0E: float,  # double

        0x09: str,
        0x0F: str,
    }

    def __init__(self, function, datatype, fi):
        self.Time = 500
        self.FIndex = fi
        self.OnChangeOnly = False
        self.Function = function
        self.PreviousValue = None
        self.Value = None

        self._dataType = None
        if datatype is not None:
            self._dataType = int(datatype, 16)

        self.DataType = None
        self.DataTypeCAN = None
        if self._dataType is not None and self._dataType in DatastorageEntry.datatype_convert:
            self.DataType = DatastorageEntry.NumToTypeFunc[self._dataType]
            self.DataTypeCAN = DatastorageEntry.datatype_convert[self._dataType]
        else:
            datatype = "None" if datatype is None else datatype
            mfl.warning(f"{__class__} Datatype: '{datatype}' is not valid.")

    def Notify(self, Value):
        self.PreviousValue = self.Value
        self.Value = Value
        changed = self.Value != self.PreviousValue
        if self.Function is not None and ((self.OnChangeOnly and changed) or self.OnChangeOnly is False):
            self.Function(self.FIndex, self.Value)

    def NativeValue(self):
        if self.Value is None or self.DataType is None:
            return None
        if self.DataType == str:
            return self.Value

        return self.DataType(self.Value)


class dsproperty(property):
    """Convenience wrapper for accessing predefined CANOpen-Objects of the datastorage."""

    def __init__(self, *args, **kwargs):
        # print(kwargs)
        self.Address = kwargs["Address"]
        self.NodeId = kwargs["NodeId"]

        args2 = (self._get, self._set, self._del, "")
        super(dsproperty, self).__init__(*args2)

    def _get(self, self2):
        no = self.NodeId(self2)
        hey = self.Address.Node(no)
        dff = storage[hey]
        return dff

    def _set(self, self2, value):
        storage[self.Address.Node(self.NodeId(self2))].Value = value

    def _del(self, self2):
        pass

    def GetAddress(self, self2):
        return self.Address.Node(self.NodeId(self2))

    # def Address(self,self2):
    #    return self.Address.Node(self.NodeId(self2))


class mfDatastorage(object):
    """Manages a list of MFX_4 device CANOpen object dictionaries (OD).

    OD are stored in the xdd format as base64 in mfxdd.
    Add device-xdds, read, write, and set notifications. 
    """

    mfx4_devices = {}
    mfx4_devices["generic"] = genericXDD
    mfx4_devices["cut communication unit"] = cutXDD
    mfx4_devices["terminalpro"] = cutXDD
    mfx4_devices["mfx_4 edi"] = cutXDD
    mfx4_devices["modospro"] = cutXDD
    mfx4_devices["mfx_4 controller"] = controllerXDD

    def __init__(self):
        self.Structure = {}
        self.mfc = None

    def AddGenericDevice(self, node):
        self.AddDeviceXdd(node, genericXDD)

    def AddDeviceByName(self, node, devicename):
        if devicename is not None:
            dev = devicename.lower().rstrip('\x00')
            if dev in mfDatastorage.mfx4_devices:
                st = mfDatastorage.mfx4_devices[dev]
                self.AddDeviceXdd(node, st)
                return True
        return False

    def AddDevice(self, dev):
        if dev is not None and dev.DeviceName.Value is not None and dev.SoftwareVersion.Value is not None:
            return self.AddDeviceByName(dev._Node, dev.DeviceName.Value)
        return False

    def AddDeviceXdd(self, node, filename):
        # root = ET.parse(filename).getroot()
        filenametext = base64.b64decode(filename)  # , validate=False)
        root = ET.fromstring(filenametext)
        obj = root.findall('CANopenObject')
        for CANopenObject in obj:
            index = int(CANopenObject.get('index'), 16)
            dataType = CANopenObject.get('dataType')
            subindex = 0
            ind = FullIndex(node, index, subindex)
            if ind not in self.Structure and dataType is not None:
                self.Structure[ind] = DatastorageEntry(self.TheVoid, dataType, ind)
                mfl.debug(f"SUCCESS-TAG: 0x{index:4X}")

            for CANopenSubObject in CANopenObject.findall('CANopenSubObject'):
                dataType2 = CANopenSubObject.get('dataType')
                subindex2 = int(CANopenSubObject.get('subIndex'), 16)
                ind2 = FullIndex(node, index, subindex2)
                if ind2 not in self.Structure and dataType2 is not None:
                    self.Structure[ind2] = DatastorageEntry(self.TheVoid, dataType2, ind2)
                mfl.debug(f"SUCCESS-SUB-TAG: 0x{index:4X} 0x{subindex2:2X}")

        mfl.debug(f"Datastorage: {len(self.Structure)} done")

    def Notify(self, FI, Value):
        if self[FI] is not None:
            self[FI].Notify(Value)
        else:
            mfl.warning(f"{FI} not in Datastorage-Structure")

    def RefreshNotification(self, FI):
        self.mfc.Send(mfx.SetNotification(FI, self[FI].Time))

    def SetNotification(self, FI, time=500, Func=None, OnChangeOnly=False):
        if isinstance(FI, dsproperty) or isinstance(FI, DatastorageEntry):
            FI = FI.FIndex
        self.mfc.Send(mfx.SetNotification(FI, time))
        self[FI].Time = time
        self[FI].OnChangeOnly = OnChangeOnly
        if Func is not None:
            self[FI].Function = Func

    def RemoveNotification(self, FI):
        if isinstance(FI, dsproperty) or isinstance(FI, DatastorageEntry):
            FI = FI.FIndex
        self.mfc.Send(mfx.RemoveNotification(FI))
        self[FI].Function = None

    def Get(self, FI, overrideCANType=None):
        if isinstance(FI, DatastorageEntry):
            FI = FI.FIndex
        if FI.NodeId == 0:
            self.mfc.Send(mfx.ReadObject(FI))
        elif overrideCANType is not None:
            self.mfc.Send(mfx.ReadObjectCan(FI, overrideCANType))
        elif self[FI].DataTypeCAN is not None:
            self.mfc.Send(mfx.ReadObjectCan(FI, self[FI].DataTypeCAN))

        if self[FI] is None:
            return None
        return self.Structure[FI].Value

    def GetNative(self, FI):
        if isinstance(FI, DatastorageEntry):
            FI = FI.FIndex
        if self[FI] is None:
            return None
        return self[FI].NativeValue()

    def Set(self, FI, Value, overrideCANType=None):
        if isinstance(FI, DatastorageEntry):
            FI = FI.FIndex
        if FI.NodeId == 0:
            self.mfc.Send(mfx.WriteObject(FI, Value))
        elif overrideCANType is not None:
            self.mfc.Send(mfx.WriteObjectCan(FI, overrideCANType, Value))
        elif self[FI].DataTypeCAN is not None:
            self.mfc.Send(mfx.WriteObjectCan(FI, self[FI].DataTypeCAN, Value))

    def SetBinary(self, FI, Value):
        self.Set(FI, Value, "b")

    def GetBinary(self, FI):
        return self.Get(FI, "b")

    def TheVoid(self, FI, Value):
        pass

    def __getitem__(self, key):
        if key in self.Structure:
            return self.Structure[key]
        mfl.warning(f"{__class__}; Key {key} not found in Datastorage.")
        return None


# TODO
# The pymfx4 package will only maintain one connection
# with a shared datastorage object defined here.
storage = mfDatastorage()
storage.AddGenericDevice(0)



from pymfx4.mffullindex import FullIndex

"""MFX_4 Protocol wrappers.

Read, Write, Re/setNotification.
"""

def LCRWrapper(func):
    """Old format with bbc."""
    def wrapper(*args, **kwargs):
        st = func(*args, **kwargs)
        byt = f"{chr(2)}0{st}{chr(3)}"
        checksum = 2
        for b in byt:
            checksum ^= ord(b)
        checksum = checksum % 128
        if checksum < 0x20:
            checksum += 0x40
        byt = byt + chr(checksum)
        return byt
    return wrapper


def WriteObject(fIndex, data):
    fIndex.NodeId = 0
    return f"WO_{0}_{fIndex.Index}_{fIndex.SubIndex}_{data}`"


def ReadObject(fIndex):
    fIndex.NodeId = 0
    return f"RO_{fIndex.NodeId}_{fIndex.Index}_{fIndex.SubIndex}`"


def WriteObjectCan(fIndex, datatype, data):
    return f"WC_{fIndex.NodeId}_{fIndex.Index}_{fIndex.SubIndex}_{datatype}_{data}`"


def ReadObjectCan(fIndex, datatype):
    return f"RC_{fIndex.NodeId}_{fIndex.Index}_{fIndex.SubIndex}_{datatype}`"


def SetNotification(fIndex: FullIndex, time=500):
    if fIndex.NodeId > 0:
        return f"SM_{fIndex.NodeId}_{fIndex.Index}_{fIndex.SubIndex}_{time}`"
    else:
        return f"SN_{fIndex.Index}_{fIndex.SubIndex}_{time}`"


def RemoveNotification(fIndex: FullIndex):
    if fIndex.NodeId > 0:
        return f"CM_{fIndex.NodeId}_{fIndex.Index}_{fIndex.SubIndex}`"
    else:
        return f"CN_{fIndex.Index}_{fIndex.SubIndex}`"



from pymfx4.mfx4ARMadilloEnums import CModuleEnum
from pymfx4.mfx4ARMadilloEnums import CLogEnum
from pymfx4.mfx4ARMadilloEnums import CInternalADCEnum
from pymfx4.mfx4ARMadilloEnums import CSystemEnum
from pymfx4.mfx4ARMadilloEnums import CClockEnum
from pymfx4.mfx4ARMadilloEnums import CAccessControlEnum
from pymfx4.mfx4ARMadilloEnums import CWatchdogEnum

from pymfx4.mfx4ARMadilloEnums import CADConverterEnum
from pymfx4.mfx4ARMadilloEnums import CDAConverterEnum

from pymfx4.mfx4ARMadilloEnums import CPIDControlEnum
from pymfx4.mfx4ARMadilloEnums import CTemperatureEnum
from pymfx4.mfx4ARMadilloEnums import CCounterEnum

from pymfx4.mfx4ARMadilloEnums import CBoardDisplayEnum

from pymfx4.mfx4ARMadilloEnums import CBoardRelayEnum
from pymfx4.mfx4ARMadilloEnums import CBoardDigitalInputEnum
from pymfx4.mfx4ARMadilloEnums import CBoardDigitalOutputEnum
from pymfx4.mfx4ARMadilloEnums import CBoardImpulseEnum
from pymfx4.mfx4ARMadilloEnums import CPulseGeneratorEnum
from pymfx4.mfx4ARMadilloEnums import CSerialEnum

from pymfx4.mfx4ARMadilloEnums import CProcessControlEnum
from pymfx4.mfx4ARMadilloEnums import CProtocolMFX4Enum
from pymfx4.mfx4ARMadilloEnums import CMMemoryEnum
from pymfx4.mfx4ARMadilloEnums import CDensityEnum
from pymfx4.mfx4ARMadilloEnums import CPressureEnum

from pymfx4.mfx4ARMadilloEnums import CErrorEnum
from pymfx4.mfx4ARMadilloEnums import CPulseEnum
from pymfx4.mfx4ARMadilloEnums import CWMHistoryEnum
from pymfx4.mfx4ARMadilloEnums import CProductEnum
from pymfx4.mfx4ARMadilloEnums import CVoucherEnum

from enum import IntEnum
import struct
import jsonpickle


class CMemoryLocations:
    """Important memory locations.

    Synchronize with ARMadillo project!
    """

    W25R_SECTOR_SIZE = 4096
    xFLASH_COPY_COUNT = 2
    xPARAMETER_MAX_SIZE = 8 * W25R_SECTOR_SIZE
    xLINKAGE_MAX_SIZE = 8 * W25R_SECTOR_SIZE
    xModuleAccessMap_MAX_SIZE = 8 * W25R_SECTOR_SIZE
    xModuleQuantityMap_MAX_SIZE = W25R_SECTOR_SIZE

    xConfigurationRAMBufferSize = 8 * W25R_SECTOR_SIZE

    xFlashParameterLocation = 0
    xFlashParameterSize = xFLASH_COPY_COUNT * xPARAMETER_MAX_SIZE
    xFlashParameterSingleSize = xPARAMETER_MAX_SIZE

    xFlashModuleLinkageLocation = xFlashParameterLocation + xFlashParameterSize
    xFlashModuleLinkageSize = xLINKAGE_MAX_SIZE

    xFlashModuleAccessLocation = xFlashModuleLinkageLocation + xFlashModuleLinkageSize
    xFlashModulAccessSize = xModuleAccessMap_MAX_SIZE

    xFlashModuleQuantityLocation = xFlashModuleAccessLocation + xFlashModulAccessSize
    xFlashModuleQuantitySize = xModuleQuantityMap_MAX_SIZE

    xFlashSystemLogsLocation = xFlashModuleQuantityLocation + xFlashModuleQuantitySize
    xFlashSystemLogsSize = 4 * W25R_SECTOR_SIZE

    xFlashLogsLocation = xFlashSystemLogsLocation + xFlashSystemLogsSize
    xFlashLogsSize = 16 * W25R_SECTOR_SIZE

    xFlashCounterProductLinkageLocation = xFlashLogsLocation + xFlashLogsSize
    xFlashCounterProductLinkageSize = W25R_SECTOR_SIZE

    xFlashProductProcessControlLinkageLocation = xFlashCounterProductLinkageLocation + xFlashCounterProductLinkageSize
    xFlashProductProcessControlLinkageSize = W25R_SECTOR_SIZE

    ulVoucherSlotSize = 1024
    ulVoucherSlots = 8
    xFlashVoucherLocation = 0x400000
    xFlashVoucherSize = ulVoucherSlots * ulVoucherSlotSize


ModuleToValues = {
    CModuleEnum.TYPE_LOG: CLogEnum,
    CModuleEnum.TYPE_INTERNAL_ADC: CInternalADCEnum,
    CModuleEnum.TYPE_SYSTEM: CSystemEnum,
    CModuleEnum.TYPE_CLOCK: CClockEnum,
    CModuleEnum.TYPE_ACCESSCONTROL: CAccessControlEnum,
    CModuleEnum.TYPE_WATCHDOG: CWatchdogEnum,

    CModuleEnum.TYPE_ADCONVERTER: CADConverterEnum,
    CModuleEnum.TYPE_DACONVERTER: CDAConverterEnum,

    CModuleEnum.TYPE_PIDCONTROL: CPIDControlEnum,
    CModuleEnum.TYPE_TEMPERATURE: CTemperatureEnum,
    CModuleEnum.TYPE_COUNTER: CCounterEnum,

    CModuleEnum.TYPE_BOARDDISPLAY: CBoardDisplayEnum,

    CModuleEnum.TYPE_BOARDRELAY: CBoardRelayEnum,
    CModuleEnum.TYPE_BOARDDIGITALINPUT: CBoardDigitalInputEnum,
    CModuleEnum.TYPE_BOARDDIGITALOUTPUT: CBoardDigitalOutputEnum,
    CModuleEnum.TYPE_BOARDIMPULSE: CBoardImpulseEnum,
    CModuleEnum.TYPE_PULSEGENERATOR: CPulseGeneratorEnum,
    CModuleEnum.TYPE_SERIAL: CSerialEnum,

    CModuleEnum.TYPE_PROCESSCONTROL: CProcessControlEnum,
    CModuleEnum.TYPE_PROTOCOL_MFX4: CProtocolMFX4Enum,
    CModuleEnum.TYPE_MEMORY_INTERFACE: CMMemoryEnum,
    CModuleEnum.TYPE_DENSITY: CDensityEnum,
    CModuleEnum.TYPE_PRESSURE: CPressureEnum,

    CModuleEnum.TYPE_ERROR: CErrorEnum,
    CModuleEnum.TYPE_PULSE: CPulseEnum,
    CModuleEnum.TYPE_WEIGHT_AND_MEASUREMENT_HISTORY: CWMHistoryEnum,
    CModuleEnum.TYPE_PRODUCT: CProductEnum,
    CModuleEnum.TYPE_VOUCHER: CVoucherEnum,
}


class commonHeader_t:
    """A header class that is common for most ARMadillo configurations.

    Imported class. Binary representation used in the ARMadillo.
    These types of configuration are using the commonHeader_t class: QuantityMap, AccessMap, LinkageMap.
    The header includes a field for the size of the whole configuration.
    """

    def __init__(self):
        self.ulLabel = 0
        self.ucCopyId = 0
        self.usOverallChecksum = 0
        self.ulTotalByteCount = 0

    def GetxNON_CRC_BYTES(self):
        """This many bytes need to be skipped for the crc calculation."""
        return 7

    def usCalculateCRC(self):
        return commonHeader_t.crc_calc(0, self.Pack())

    def xCheckCRC(self):
        return self.usOverallChecksum == self.usCalculateCRC()

    def vCalculateAndSetCRC(self):
        self.usOverallChecksum = self.usCalculateCRC()

    def crc_calc(crc, data):
        crc &= 0xffff

        header = commonHeader_t()
        header.Unpack(data)

        # uint16_t
        usTable = [0] * 256

        for i in range(0, 256):
            k = 0xC0C1
            usTable[i] = 0
            for j in [1, 2, 4, 8, 16, 32, 64, 128]:
                if (i & j) > 0:
                    usTable[i] ^= k
                    usTable[i] &= 0xffff
                k = ((k << 1) ^ 0x4003)

        dat = data[header.GetxNON_CRC_BYTES():header.ulTotalByteCount]
        for p in dat:
            crc = (crc >> 8) ^ usTable[(crc & 0xFF) ^ p]

        return crc

    def Pack(self):
        return struct.pack('<IBHI', self.ulLabel, self.ucCopyId, self.usOverallChecksum, self.ulTotalByteCount)

    def Unpack(self, data):
        d = data[:11]
        self.ulLabel, self.ucCopyId, self.usOverallChecksum, self.ulTotalByteCount = struct.unpack('<IBHI', d)

    def __repr__(self):
        options = {moduleAccessMapHeader_t.ulMODULE_ACCESS_MAP_LABEL: "Module Access Map",
                   moduleLinkageHeader_t.ulMODULE_LINKAGE_LABEL: "Module Linkage Map",
                   moduleQuantityHeader_t.ulMODULE_QUANTITY_LABEL: "Module Quantity Map"}
        header = "Configuration?"
        if self.ulLabel in options.keys():
            header = options[self.ulLabel]
        header = f"{header} ({self.ucCopyId})\n"
        header += f"CRC: 0x{self.usOverallChecksum:X}\n"
        header += f"Bytes: {self.ulTotalByteCount}\n"

        return f'{header}'


class moduleId_t:
    """Representing a ARMadillo module with module type and sequence number.

    Imported class. Binary representation used in the ARMadillo.
    """

    def G(usSeqNumArg, xTypeArg):
        a = moduleId_t()
        a.usSeqNum = usSeqNumArg
        try:
            a.xType = CModuleEnum(xTypeArg)
        except:
            a.xType = CModuleEnum.TYPE_MAX
        return a

    def __init__(self):
        self.usSeqNum = 0
        self.xType = CModuleEnum.TYPE_MAX

    def __repr__(self):
        usSeq = f'{self.usSeqNum:02d}'
        if self.usSeqNum == 0xffff:
            usSeq = "?"

        xT = f'{self.xType:02d}'
        name = self.xType.name
        if int(self.xType) == 0xff:
            xT = "?"
            name = "?"

        return f'[{usSeq}][{xT}] {name}'

    def Pack(self):
        return struct.pack('<HH', self.usSeqNum, self.xType)


class uniqueValueId_t:
    """Representing an address to a ARMadillo-modules value-container.

    Imported class. Binary representation used in the ARMadillo.
    """

    def G(xModuleIdArg, xValueIdArg):
        a = uniqueValueId_t()
        a.xModuleId = xModuleIdArg
        a.xValueId = xValueIdArg
        return a

    def __init__(self):
        self.xModuleId = 0
        self.xValueId = 0

    def Pack(self):
        return self.xModuleId.Pack() + struct.pack('<H', self.xValueId)

    def Unpack(self, data):
        d = data[:6]
        a, b, self.xValueId, = struct.unpack('<HHH', d)
        self.xModuleId = moduleId_t.G(a, b)

    def __repr__(self):
        ret = f'*{self.xValueId}'
        if self.xModuleId.xType in ModuleToValues.keys():
            values = [item.value for item in ModuleToValues[self.xModuleId.xType]]
            if self.xValueId in values:
                ret = ModuleToValues[self.xModuleId.xType](self.xValueId)
                ret = f'{ret.name} *{ret.value}'

        return f'{self.xModuleId}=={ret}'


class moduleConfigurationHeader:
    """Base class of most ARMadillo configrations.

    Imported class. Binary representation used in the ARMadillo.
    Helper functions include import and export.
    """

    def __init__(self):
        self.xCommonHeader = commonHeader_t()

    def Prepare(self):
        self.xCommonHeader.ulTotalByteCount = len(self.Pack())
        self.xCommonHeader.usOverallChecksum = commonHeader_t.crc_calc(0, self.Pack())
        pass

    def Export(obj, filePath=None):
        jsondata = jsonpickle.encode(obj, make_refs=False, indent=3)

        import os
        namespace = "pymfx4." + os.path.basename(__file__).replace(".py", "")

        jsondata = jsondata.replace("__main__", namespace)

        # jsondata = pickle.dumps(obj)
        if filePath:
            try:
                with open(filePath, 'wt') as f:
                    f.write(jsondata)
            except:
                pass
        return jsondata

    def Import(filePath=None, data=None):
        if filePath:
            with open(filePath, 'rt') as f:
                data = f.read()
        if data:
            return jsonpickle.decode(data, safe=True)  # , classes=[moduleQuantityHeader_t, moduleQuantity_t, CModuleEnum])
        return None


class moduleLinkageHeader_t(moduleConfigurationHeader):
    """Module linkage ARMadillo configuration.

    Imported class. Binary representation used in the ARMadillo.
    """

    ulMODULE_LINKAGE_LABEL = 0x4D4C0000

    def __init__(self):
        super().__init__()
        self.xCommonHeader.ulLabel = moduleLinkageHeader_t.ulMODULE_LINKAGE_LABEL
        self.ulModuleLinkageMappingCount = 0
        self.xMappings = []
        self.Prepare()

    def Pack(self):
        ret = self.xCommonHeader.Pack() + struct.pack('<I', self.ulModuleLinkageMappingCount)
        for m in self.xMappings:
            ret += m.Pack()
        return ret

    def Unpack(self, data):
        self.xCommonHeader.Unpack(data)
        d = data[11:15]
        self.ulModuleLinkageMappingCount, = struct.unpack('<I', d)
        d = data[15:]
        for i in range(self.ulModuleLinkageMappingCount):
            mapping = moduleLinkageMapping_t()
            mapping.Unpack(d)
            self.xMappings.append(mapping)
            d = d[7 + mapping.xConsumerCount * 6:]

    def __repr__(self):
        ret = f'{self.xCommonHeader}'
        for i in self.xMappings:
            ret += f"{i}"
        return ret

    def Sort(self):
        # def index(elem):
        #     return elem.xProducerId.xModuleId.xType
        # self.xMappings.sort(key=index)

        self.xMappings.sort(key=lambda x: x.xProducerId.xModuleId.xType)

    def Prepare(self):
        self.ulModuleLinkageMappingCount = len(self.xMappings)
        super().Prepare()
        pass


class moduleLinkageMapping_t:
    """Module linkage ARMadillo configuration content.

    Imported class. Binary representation used in the ARMadillo.
    """

    def G(xProducerIdArg, *pxConsumerIds):
        a = moduleLinkageMapping_t()
        a.xConsumerCount = len(pxConsumerIds)
        a.xProducerId = xProducerIdArg

        a.pxConsumerIds = []
        for arg in pxConsumerIds:
            a.pxConsumerIds.append(arg)
        return a

    def __init__(self):
        self.xConsumerCount = 0
        self.xProducerId = 0
        self.pxConsumerIds = []

    def Pack(self):
        ret = struct.pack('<B', self.xConsumerCount)
        ret += self.xProducerId.Pack()
        for ids in self.pxConsumerIds:
            ret += ids.Pack()
        return ret

    def Unpack(self, data):
        self.xConsumerCount, = struct.unpack('<B', data[0:1])
        self.xProducerId = uniqueValueId_t()
        self.xProducerId.Unpack(data[1:7])
        for i in range(self.xConsumerCount):
            uVId = uniqueValueId_t()
            uVId.Unpack(data[7 + (i * 6):7 + (i * 6 + 6)])
            self.pxConsumerIds.append(uVId)

    def __repr__(self):
        ret = f'{self.xProducerId} =>\n'
        for i in self.pxConsumerIds:
            ret += f'\t{i}\n'
        return ret


class moduleAccessMapHeader_t(moduleConfigurationHeader):
    """Module access ARMadillo configuration.

    Imported class. Binary representation used in the ARMadillo.
    """

    ulMODULE_ACCESS_MAP_LABEL = 0x434f494d

    def __init__(self):
        super().__init__()
        self.xCommonHeader.ulLabel = moduleAccessMapHeader_t.ulMODULE_ACCESS_MAP_LABEL
        self.moduleAccessObjects = []
        self.Prepare()

    def Unpack(self, data):
        self.xCommonHeader.Unpack(data)
        d = data[11:]
        parsed = 11
        while parsed < self.xCommonHeader.ulTotalByteCount:
            obj = moduleAccessObjectHeader_t()
            obj.usIndex, obj.ucSubCount, = struct.unpack('<HB', d[:3])

            d = d[3:]
            parsed += 3
            for i in range(obj.ucSubCount + 1):
                sub = moduleAccessSubObject_t()
                sub.Unpack(d)
                obj.pxSubObjects.append(sub)
                d = d[11:]
                parsed += 11

            self.moduleAccessObjects.append(obj)

    def Pack(self):
        ret = self.xCommonHeader.Pack()
        for m in self.moduleAccessObjects:
            ret += m.Pack()
        return ret

    def __repr__(self):
        ret = f'{self.xCommonHeader}'

        total = 0
        for i in self.moduleAccessObjects:
            total += len(i.pxSubObjects)

        ret += f'Total SubObjects: {total}\n'

        for i in self.moduleAccessObjects:
            ret += f'\t{i}'
        return ret

    def Prepare(self):
        for i in self.moduleAccessObjects:
            i.ucSubCount = len(i.pxSubObjects) - 1
            if len(i.pxSubObjects) - 1 < 0:
                i.ucSubCount = 0
        self.Sort()
        super().Prepare()
        pass

    def Sort(self):
        # self.moduleAccessObjects.sort(key=index)
        # self.moduleAccessObjects.sort(key=lambda x: x.usIndex)
        for i in self.moduleAccessObjects:
            # i.pxSubObjects.sort(key=subindex)
            # i.pxSubObjects.sort(key=lambda x: x.xUniqueValueId.xValueId)
            i.pxSubObjects.sort(key=lambda x: x.ucSubIndex)
            i.ucSubCount = len(i.pxSubObjects) - 1
            if i.ucSubCount < 0:
                i.ucSubCount = 0


class moduleAccessObjectHeader_t:
    """Module access ARMadillo configuration content.

    Imported class. Binary representation used in the ARMadillo.
    """

    def __init__(self):
        self.usIndex = 0
        self.ucSubCount = 0
        self.pxSubObjects = []

    def __repr__(self):
        ret = f'Object 0x{self.usIndex:X} (Count:{self.ucSubCount})\n'
        for i in self.pxSubObjects:
            ret += f'\t\t{i}\n'
        return ret

    def Pack(self):
        ret = struct.pack('<HB', self.usIndex, self.ucSubCount)
        for m in self.pxSubObjects:
            ret += m.Pack()
        return ret


class moduleAccessSubObject_t:
    """Module access ARMadillo configuration content.

    Imported class. Binary representation used in the ARMadillo.
    """

    def __init__(self):
        self.ucSubIndex = 0
        self.usByteCount = 0
        self.xUniqueValueId = uniqueValueId_t()
        self.ucFlags = 0
        self.eMFXType = 0  # byte

    def Unpack(self, data):
        self.ucSubIndex, self.usByteCount, = struct.unpack('<BH', data[:3])
        self.xUniqueValueId.Unpack(data[3:])
        self.ucFlags, self.eMFXType, = struct.unpack('<BB', data[9:11])

    def Pack(self):
        ret = struct.pack('<BH', self.ucSubIndex, self.usByteCount)
        ret += self.xUniqueValueId.Pack()
        ret += struct.pack('<BB', self.ucFlags, self.eMFXType)
        return ret

    def __repr__(self):
        return f"{self.xUniqueValueId} (Sub:0x{self.ucSubIndex:X}/{self.ucSubIndex}) (Bytes:{self.usByteCount})"


class moduleQuantityHeader_t(moduleConfigurationHeader):
    """Module quantity ARMadillo configuration content.

    Imported class. Binary representation used in the ARMadillo.
    """

    ulMODULE_QUANTITY_LABEL = 0x514d0000

    def __init__(self):
        super().__init__()
        self.xCommonHeader.ulLabel = moduleQuantityHeader_t.ulMODULE_QUANTITY_LABEL
        self.usModuleCount = 0
        self.pxModuleQuantities = []
        self.Prepare()

    def Unpack(self, data):
        self.xCommonHeader.Unpack(data)
        d = data[11:]
        self.usModuleCount, = struct.unpack('<H', d[:2])
        d = d[2:]
        for i in range(self.usModuleCount):
            obj = moduleQuantity_t()
            obj.Unpack(d[:4])
            self.pxModuleQuantities.append(obj)
            d = d[4:]

    def Pack(self):
        ret = self.xCommonHeader.Pack()
        ret += struct.pack("<H", self.usModuleCount)
        for m in self.pxModuleQuantities:
            ret += m.Pack()
        return ret

    def __repr__(self):
        ret = f'{self.xCommonHeader}'
        ret += f'Entries: {self.usModuleCount}\n'
        for i in self.pxModuleQuantities:
            ret += f'\t{i}\n'
        return ret

    def Sort(self):
        self.pxModuleQuantities.sort(key=lambda x: x.xModuleType)
        super().Prepare()

    def Prepare(self):
        self.Sort()
        self.usModuleCount = len(self.pxModuleQuantities)
        super().Prepare()
        pass


class moduleQuantity_t:
    """Module quantity ARMadillo configuration content.

    Imported class. Binary representation used in the ARMadillo.
    """
    def G(xModuleType, usModuleQuantity):
        quan = moduleQuantity_t()
        quan.xModuleType = xModuleType
        quan.usModuleQuantity = usModuleQuantity
        return quan

    def __init__(self):
        self.xModuleType = CModuleEnum.TYPE_MAX
        self.usModuleQuantity = 0

    def Unpack(self, data):
        a, self.usModuleQuantity = struct.unpack('<HH', data[:4])
        values = set(item.value for item in CModuleEnum)
        if a in values:
            self.xModuleType = CModuleEnum(a)

    def Pack(self):
        return struct.pack('<HH', self.xModuleType, self.usModuleQuantity)

    def __repr__(self):
        return f"{self.xModuleType.name}: {self.usModuleQuantity}"


class logSeverity_t(IntEnum):
    """Severity of reported log content.

    Imported class. Binary representation used in the ARMadillo.
    """

    NOMASK = 0x30,
    ERROR = 0x31
    WARNING = 0x32
    HINT = 0x33


class logMessage_t:
    """Representation of an ARMadillo log message .

    Imported class. Binary representation used in the ARMadillo.
    """

    LOG_MAX_SOURCEFILE_LEN = 10
    LOG_MAX_COMMENT_LEN = 37

    def __init__(self):
        self.xTimestamp = 0  # 8byte
        self.xSeverity = logSeverity_t.NOMASK.value  # 1byte
        self.xModuleId = moduleId_t()
        self.pcSourceFile = ""  # LOG_MAX_SOURCEFILE_LEN
        self.usLineNumber = 0  # 2byte
        self.usCause = 0   # 2byte
        self.pcComment = ""  # LOG_MAX_COMMENT_LEN

    def Unpack(self, data):
        self.xTimestamp, self.xSeverity, a, b, self.pcSourceFile, self.usLineNumber, self.usCause, self.pcComment = struct.unpack(f'<QBHH{logMessage_t.LOG_MAX_SOURCEFILE_LEN}sHH{logMessage_t.LOG_MAX_COMMENT_LEN}s', data[:64])
        self.xModuleId = moduleId_t.G(a, b)

    def Pack(self):
        ret = struct.pack('<QB', self.xTimestamp, self.xSeverity)
        ret += self.xModuleId.Pack()
        sourceFileString = f"{self.pcSourceFile:<{logMessage_t.LOG_MAX_SOURCEFILE_LEN}}"
        if len(sourceFileString) > logMessage_t.LOG_MAX_SOURCEFILE_LEN:
            sourceFileString = sourceFileString[0:logMessage_t.LOG_MAX_SOURCEFILE_LEN]
        ret += struct.pack(f'<{logMessage_t.LOG_MAX_SOURCEFILE_LEN}sHH', bytes(sourceFileString, encoding="ASCII"), self.usLineNumber, self.usCause)

        comment = f"{self.pcComment:<{logMessage_t.LOG_MAX_COMMENT_LEN}}"
        if len(comment) > logMessage_t.LOG_MAX_COMMENT_LEN:
            comment = comment[0:logMessage_t.LOG_MAX_COMMENT_LEN]
        ret += struct.pack('<{logMessage_t.LOG_MAX_COMMENT_LEN}s', bytes(comment, encoding="ASCII"))
        return ret

    def __repr__(self):
        a = self.pcSourceFile.encode("ascii")
        b = self.pcComment.encode("ascii")
        return f"{self.xTimestamp :<10} {self.xModuleId}: {a} {b}"


class ErrorConfiguration_t:
    """Representation of an ARMadillo error configuration.
    
    Imported class. Binary representation used in the ARMadillo.
    [todo]
    """

    ERROR_NO_BIT = 0
    ERROR_NO_VALUE = 0x7FF << ERROR_NO_BIT
    ERRORLESS_PART_GENERATION_BIT = 11
    ERRORLESS_PART_GENERATION_VALUE = 0x1 << ERRORLESS_PART_GENERATION_BIT
    CAUSES_BIT = 12
    CAUSES_VALUE = 0x7 << CAUSES_BIT
    RESET_BY_COMMAND_BIT = 15
    RESET_BY_COMMAND_VALUE = 0x1 << 15
    DELAY_BREAK_TIME_BIT = 16
    DELAY_BREAK_TIME_VALUE = 0xFFF << DELAY_BREAK_TIME_BIT
    OUTPUT_BIT = 28
    OUTPUT_VALUE = 0xf << OUTPUT_BIT

    def __init__(self):
        self.ulErrorProperties = 0  # 4byte
        self.ulPermissionProperties = 0  # 4byte
        self.ulProcessControls = 0  # 4byte

    def Unpack(self, data):
        self.ulErrorProperties, self.ulPermissionProperties, self.ulProcessControls = struct.unpack('<III', data[:12])

    def Pack(self):
        ret = struct.pack('<III', self.ulErrorProperties, self.ulPermissionProperties, self.ulProcessControls)
        return ret

    def __repr__(self):
        # ErrorClass = (self.ulErrorProperties & ErrorConfiguration_t.ERROR_NO_VALUE) >> ErrorConfiguration_t.ERROR_NO_BIT
        # Cause = (self.ulErrorProperties & ErrorConfiguration_t.CAUSES_VALUE) >> ErrorConfiguration_t.CAUSES_BIT
        # ResetCommand = (self.ulErrorProperties & ErrorConfiguration_t.RESET_BY_COMMAND_VALUE) >> ErrorConfiguration_t.RESET_BY_COMMAND_BIT
        # DelayBreakTime = (self.ulErrorProperties & ErrorConfiguration_t.DELAY_BREAK_TIME_VALUE) >> ErrorConfiguration_t.DELAY_BREAK_TIME_BIT
        # Output = (self.ulErrorProperties & ErrorConfiguration_t.OUTPUT_VALUE) >> ErrorConfiguration_t.OUTPUT_BIT

        return f"EC{self.ErrorClass:<5} F{self.GenerateErrorlessPart} C{self.Cause} RES{self.ResetCommand} DELAY{self.DelayBreakTime} OUT{self.Output}"

    @property
    def GenerateErrorlessPart(self):
        return (self.ulErrorProperties & ErrorConfiguration_t.ERRORLESS_PART_GENERATION_VALUE) >> ErrorConfiguration_t.ERRORLESS_PART_GENERATION_BIT

    @GenerateErrorlessPart.setter
    def GenerateErrorlessPart(self, x):
        self.ulErrorProperties &= ~ErrorConfiguration_t.ERRORLESS_PART_GENERATION_VALUE
        x = x & (ErrorConfiguration_t.ERRORLESS_PART_GENERATION_VALUE >> ErrorConfiguration_t.ERRORLESS_PART_GENERATION_BIT)
        self.ulErrorProperties |= x << ErrorConfiguration_t.ERRORLESS_PART_GENERATION_BIT

    @property
    def Output(self):
        return (self.ulErrorProperties & ErrorConfiguration_t.OUTPUT_VALUE) >> ErrorConfiguration_t.OUTPUT_BIT

    @Output.setter
    def Output(self, x):
        self.ulErrorProperties &= ~ErrorConfiguration_t.OUTPUT_VALUE
        x = x & (ErrorConfiguration_t.OUTPUT_VALUE >> ErrorConfiguration_t.OUTPUT_BIT)
        self.ulErrorProperties |= x << ErrorConfiguration_t.OUTPUT_BIT

    @property
    def DelayBreakTime(self):
        return (self.ulErrorProperties & ErrorConfiguration_t.DELAY_BREAK_TIME_VALUE) >> ErrorConfiguration_t.DELAY_BREAK_TIME_BIT

    @DelayBreakTime.setter
    def DelayBreakTime(self, x):
        self.ulErrorProperties &= ~ErrorConfiguration_t.DELAY_BREAK_TIME_VALUE
        x = x & (ErrorConfiguration_t.DELAY_BREAK_TIME_VALUE >> ErrorConfiguration_t.DELAY_BREAK_TIME_BIT)
        self.ulErrorProperties |= x << ErrorConfiguration_t.DELAY_BREAK_TIME_BIT

    @property
    def ResetCommand(self):
        return (self.ulErrorProperties & ErrorConfiguration_t.RESET_BY_COMMAND_VALUE) >> ErrorConfiguration_t.RESET_BY_COMMAND_BIT

    @ResetCommand.setter
    def ResetCommand(self, x):
        self.ulErrorProperties &= ~ErrorConfiguration_t.RESET_BY_COMMAND_VALUE
        x = x & (ErrorConfiguration_t.RESET_BY_COMMAND_VALUE >> ErrorConfiguration_t.RESET_BY_COMMAND_BIT)
        self.ulErrorProperties |= x << ErrorConfiguration_t.RESET_BY_COMMAND_BIT

    @property
    def Cause(self):
        return (self.ulErrorProperties & ErrorConfiguration_t.CAUSES_VALUE) >> ErrorConfiguration_t.CAUSES_BIT

    @Cause.setter
    def Cause(self, x):
        self.ulErrorProperties &= ~ErrorConfiguration_t.CAUSES_VALUE
        x = x & (ErrorConfiguration_t.CAUSES_VALUE >> ErrorConfiguration_t.CAUSES_BIT)
        self.ulErrorProperties |= x << ErrorConfiguration_t.CAUSES_BIT

    @property
    def ErrorClass(self):
        return (self.ulErrorProperties & ErrorConfiguration_t.ERROR_NO_VALUE) >> ErrorConfiguration_t.ERROR_NO_BIT

    @ErrorClass.setter
    def ErrorClass(self, x):
        self.ulErrorProperties &= ~ErrorConfiguration_t.ERROR_NO_VALUE
        x = x & (ErrorConfiguration_t.ERROR_NO_VALUE >> ErrorConfiguration_t.ERROR_NO_BIT)
        self.ulErrorProperties |= x << ErrorConfiguration_t.ERROR_NO_BIT



"""Provides aliases, attributes and functions for simulating key-press events in the MFX_4 Terminal."""

class keyboard:
    """Provides aliases, attributes and functions for simulating key-press events in the MFX_4 Terminal."""

    def InputNumber(number):
        lst = []
        for n in str(number):
            i = ord(n) - ord('0')
        v = keyboard.Numbers[i]
        lst.append(v)
        return lst

    F1 = 91
    F2 = 92
    F3 = 93
    F4 = 94

    F1 = 1
    F2 = 2
    F3 = 3
    F4 = 4

    # Numbers = [47, 67, 68, 69, 75, 76, 77, 83, 84, 85]
    Numbers = [0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39]

    Null = Numbers[0]
    Eins = Numbers[1]
    Zwei = Numbers[2]
    Drei = Numbers[3]
    Vier = Numbers[4]
    Fuenf = Numbers[5]
    Sechs = Numbers[6]
    Sieben = Numbers[7]
    Acht = Numbers[8]
    Neun = Numbers[9]

    Alphabet = [65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88]
    A = Alphabet[0]
    B = Alphabet[1]
    C = Alphabet[2]
    D = Alphabet[3]
    E = Alphabet[4]
    F = Alphabet[5]
    G = Alphabet[6]
    H = Alphabet[7]
    I = Alphabet[8]
    J = Alphabet[9]
    K = Alphabet[10]
    L = Alphabet[11]
    M = Alphabet[12]
    N = Alphabet[13]
    O = Alphabet[14]
    P = Alphabet[15]
    Q = Alphabet[16]
    R = Alphabet[17]
    S = Alphabet[18]
    T = Alphabet[19]
    U = Alphabet[20]
    V = Alphabet[21]
    W = Alphabet[22]
    X = Alphabet[23]

    DEL = 8

    PRINT = 95
    LEFT = 89
    UP = 90
    RIGHT = 45
    DOWN = 42
    BLANK = 32
    SPECIALCHAR = 32
    STOP = 13

    # keys_all =[item for sublist in [F1,F2,F3,F4,Numbers,Alphabet,PRINT,LEFT,UP,RIGHT,DOWN,BLANK,SPECIALCHAR,STOP] for item in sublist]  #flatten list
    # keys_fkeysandnumeric  = [item for sublist in [F1,F2,F3,F4,Numbers] for item in sublist]  #flatten list
    KeysToDesc = {
        F1: "F1",
        F2: "F2",
        F3: "F3",
        F4: "F4",

        A: "A",
        B: "B",
        Numbers[1]: "1_C",
        Numbers[2]: "2_D",
        Numbers[3]: "3_E",
        F: "F",
        G: "G",
        H: "H",
        I: "I",
        J: "J",
        Numbers[4]: "4_K",
        Numbers[5]: "5_L",
        Numbers[6]: "6_M",
        N: "N",
        O: "O",
        P: "P",
        Q: "Q",
        R: "R",
        Numbers[7]: "7_S",
        Numbers[8]: "8_T",
        Numbers[9]: "9_U",
        V: "V",
        W: "W",
        X: "X",
        PRINT: "PRINT",
        LEFT: "LEFT",
        UP: "UP",
        Numbers[0]: "0",
        DOWN: "DOWN",
        RIGHT: "RIGHT",
        SPECIALCHAR: "SPECIALCHAR_BLANK",
        STOP: "STOP"
    }

    thelist = []

    @staticmethod
    def getKeyFromIndex(keyindex):
        return keyboard.keys[keyindex]

    @staticmethod
    def getDescFromKey(keybyte):
        try:
            ord(keybyte)
            return keyboard.KeysToDesc[keybyte]
        except:
            pass
        return keyboard.KeysToDesc[bytes([keybyte])]



from enum import IntEnum


class TerminalInputState(IntEnum):
    NONE = 0
    Password = 1
    Transaction = 8
    Batch = 10
    Parameter_Item = 20
    Barrels = 37
    License_Plate = 80
    Order_Number = 100
    Unknown = 255


class TerminalDisplayState(IntEnum):
    Startup = 0
    Meterlist = 1
    Loading = 3
    LoadingBlending = 5
    Session = 9
    Parameters = 10
    Clock = 40
    ErrorHistory = 45
    Language = 50
    Splash = 55
    Vouchers = 60
    Unknown = 255


class ProcessControlState(IntEnum):
    Blocked = -2
    NoStateAvailable = -1
    SystemStart = 0
    NoPara = 1
    NotReady = 2
    Idle = 3
    TransactionQuantityInput = 4
    BatchQuantityInput = 5
    VolumeWithoutOrder = 6
    ReadyToStart = 7
    Loading = 8
    Stopped = 9
    TransactionQuantityReached = 10
    LoadingInterrupted = 11
    LoadingAborted = 12
    Printing = 15
    Service = 18
    DataTransfer = 20


class ProcessControlCommand(IntEnum):
    STOP = 0
    START = 1
    RESET = 2
    SET_PRESET = 3
    PART_RESET = 4
    SET_TRANSACTION = 5
    ACK_HOST = 10
    ACK_PRINTER = 11


class BoldataState(IntEnum):
    NONE = 0
    FailedControllerOffline = 0x80
    ReqPart1 = 0x02
    FailedReqPart1CRC = 0x82
    ReqPart2 = 0x03
    FailedReqPart2CRC = 0x83
    ReqPart3 = 0x04
    FailedReqPart3CRC = 0x84
    SanityCheckWithControllerData = 0x06
    FailedSanityCheck = 0x86
    SavingLocalCopy = 0x07
    FailedSavingLocalCopy = 0x87
    FormatLines = 0x08
    FailedFormatLines = 0x88
    BuildBol = 0x09
    FailedBuildBol = 0x89
    AcknowledgeDatatransferInDataTransfer = 0x0B
    FailedAcknowledgeDatatransferInDataTransfer = 0x8B
    Failed = 0x0E,
    FailedWithException = 0xEF,
    Success = 0x01


class CARDREADERSTATE_CANOPEN(IntEnum):
    NoCardDetected = 0
    CardPresentOK = 0x1
    CardPresentBAD = 0x2
    CardPresentBeingProcessed = 0x3
    CardPresentMifareSession = 0x4
    CardPresentMifareAuth = 0x5
    CardPresentMifareReaddata = 0x06
    CardPresentMifareAuthError = 0x7
    CardPresentMifareSessionError = 0x08



# -*- coding: utf-8 -*-

from .context import pymfx4

import unittest

class BasicTestSuite(unittest.TestCase):
    """Basic test cases."""

    def test_fullindex(self):
        pass
        
    def test_keyboard(self):
        pass


if __name__ == '__main__':
    unittest.main()


import logging
logging.basicConfig(format='%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s',
                datefmt='%Y-%m-%d:%H:%M:%S',
                level=logging.ERROR)
mfl = logging.getLogger("pymfx4")
mfl.setLevel(logging.ERROR)
#ch = logging.StreamHandler()
#ch.setLevel(logging.WARNING)
#form = logging.Formatter('%(asctime)s,%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s')
#mfl setFormatter()





# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open('README.md') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

setup(
    name='pymfx4',
    version='0.1.1',
    description='Control of MFX_4 Series Devices',
    long_description=readme,
    author='Alexander Angold',
    author_email='pymfx4@infodb.eu',
    url='https://www.m-f.tech/',
    license=license,
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
          'serial>=0.0.97',
          'jsonpickle'
      ],
    test_suite='nose.collector',
    tests_require=['nose', 'nose-cover3'],
)



import pymfx4.mfdatastorage as ds
from pymfx4.mffullindex import FullIndex
from pymfx4 import mfl

import time
import re



class mfparameters:
    """A list of parameters to be send.

    Loads a list of parameters from a textfile usually generated by the MFX_4 Explorer."""

    def __init__(self, config, tim=5):
        """Loads a list of parameters from a textfile <config>."""
        self.config = config
        self.Parameters = {}
        self.load_parameters()

    def load_parameters(self):
        mfl.info(f"Loading {self.config} as Script-File.")
        line = 0
        for x in [line.rstrip('\n') for line in open(self.config)]:
            matchObj = re.match(r'(?P<cmd>.*?)_(?P<nodeid>.*?)_(?P<index>.*?)_(?P<subindex>.*?)_(?P<content>.*)`(?P<description>.*)', x, re.I | re.DOTALL)
            if matchObj:
                fi = FullIndex(int(matchObj.group("nodeid")), int(matchObj.group("index")), int(matchObj.group("subindex")))
                content = matchObj.group("content")
                description = matchObj.group("description").strip()
                fi.NodeId = 0
                self.Parameters[fi] = [content, description]
            else:
                mfl.info(f"Could not load line:{line}.")
            line += 1
        pass


class mfParameterManager:
    """Static class to read / write and check mfparameters."""

    def Check(Package, Endpoint=None):
        Node = Endpoint._Node if Endpoint is not None else 0
        for entry in Package.Parameters:
            e = entry.Node(Node)
            ds.storage.Get(e)
            mfl.info(f"get {e}")
            time.sleep(0.1)

        time.sleep(0.5)
        OKn = 0
        FAILn = 0
        for entry in Package.Parameters:

            e = entry.Node(Node)

            soll_val = Package.Parameters[entry][0] if Package.Parameters[entry][0] is None else Package.Parameters[entry][0].lower().rstrip("\x00")
            ist_val = ds.storage.Structure[e].Value if ds.storage.Structure[e].Value is None else ds.storage.Structure[e].Value.lower().rstrip("\x00")

            if ds.storage.Structure[e].DataType is None:
                mfl.error(f"{e} has no DataType ")

            soll_native_val = None if soll_val is None or ds.storage.Structure[e].DataType is None else ds.storage.Structure[e].DataType(soll_val)
            ist_native_val = ds.storage.Structure[e].NativeValue()

            oOK = (soll_val == ist_val or soll_native_val == ist_native_val) and soll_val is not None and ist_val is not None

            # stringOK = "OK"
            if oOK:
                OKn = OKn + 1
            else:
                # stringOK = "FAIL"
                FAILn = FAILn + 1

            # soll = "<None>" if Package.Parameters[entry][0] is None else soll_val
            # ist = "<None>" if ds.storage.Structure[e].Value is None else ist_val

            if ist_native_val is not None and soll_native_val is not None:
                mfl.info("{stringOK:5} {e} {soll_native_val:>20} == {ist_native_val:>20}")
            else:
                mfl.info("{stringOK:5} {e} {soll:>20} == {ist:20}")

        mfl.info(" OK:{OKn:4}  FAILED:{FAILn:4} ")
        return mfParameterManager.CheckOK((OKn, FAILn))

    def Deploy(Package, Endpoint=None):
        Node = Endpoint._Node if Endpoint is not None else 0
        for entry in Package.Parameters:
            e = entry.Node(Node)
            ds.storage.Set(e, Package.Parameters[entry][0])
            mfl.info(f"set {e}")

    def DeployAndCheck(Package, Endpoint=None):
        # Node = Endpoint._Node if Endpoint != none else 0
        mfParameterManager.Deploy(Package, Endpoint)
        time.sleep(1)
        return mfParameterManager.Check(Package, Endpoint)

    def DeployAndCheckEntrywise(Package, Endpoint=None):
        Node = Endpoint._Node if Endpoint is not None else 0
        OKn = 0
        FAILn = 0
        for entry in Package.Parameters:
            e = entry.Node(Node)
            ds.storage.Get(e)
            mfl.info(f"get {e}")
            time.sleep(0.1)

            ds.storage.Set(e, Package.Parameters[entry][0])
            mfl.info(f"set {e}")
            time.sleep(3)

            ds.storage.Get(e)
            mfl.info(f"get {e}")
            time.sleep(0.1)

            soll_val = Package.Parameters[entry][0] if Package.Parameters[entry][0] is None else Package.Parameters[entry][0].lower().rstrip("\x00")
            ist_val = ds.storage.Structure[e].Value if ds.storage.Structure[e].Value is None else ds.storage.Structure[e].Value.lower().rstrip("\x00")

            if ds.storage.Structure[e].DataType is None:
                mfl.error(f"{e} has no DataType ")

            soll_native_val = None if soll_val is None or ds.storage.Structure[e].DataType is None else ds.storage.Structure[e].DataType(soll_val)
            ist_native_val = ds.storage.Structure[e].NativeValue()

            oOK = (soll_val == ist_val or soll_native_val == ist_native_val) and soll_val is not None and ist_val is not None

            # stringOK = "OK"
            if oOK:
                OKn = OKn + 1
            else:
                # stringOK = "FAIL"
                FAILn = FAILn + 1

            # soll = "<None>" if Package.Parameters[entry][0] is None else soll_val
            # ist = "<None>" if ds.storage.Structure[e].Value is None else ist_val

            if ist_native_val is not None and soll_native_val is not None:
                mfl.info("{stringOK:5} {e} {soll_native_val:>20} == {ist_native_val:<20}")
            else:
                mfl.info("{stringOK:5} {e} {soll:>20} == {ist:<20}")

        mfl.info("RESULT OK:{OKn:4}  FAILED:{FAILn:4} ")
        return mfParameterManager.CheckOK((OKn, FAILn))

    def CheckOK(val):
        if val is None:
            return None
        ok, fail = val
        return fail == 0
